/***************************************************************************
 *   Copyright (C) 2008 by Paul Schneider   *
 *   paul.schneider@wu-wien.ac.at   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <cstdlib>
#include <ctime>
#include <fstream>


#include "utils.hpp"

// this file computes drift estimates for the square-root diffusion with representation
// dx = (a + b x_t) dt + sigma sqrt(x_t) dW_t
// using expected likelihood estimation described in Mijatovic and Schneider (2007)
// drift coefficients are computed *conditional* on sigma. To determine sigma the functions below can be passed
// to a standard gradient optimizer

// function CIRestimate returns value proportional to the likelihood of the observed time-series conditional on the square-root model
// computed with Simulated Density Approach using importance sampler from Durham and Gallant (2007)
// EML estimates for drift coefficient are written into a and b. this is why a and b are passed by reference



double CIRestimate(double* randomNumbers, // vector (of size ((DATASIZE-1) * (AUXPOINTS + 1) + 1) * CROSSSEC) containing N(0,1) random variates
				   double* data, // vector (of size dataSize) containing the data
				 	double delta, // time between observations in vector data (assumed constant for simpl.)
				 	double sigma, // value for diffusion coefficient in the SDE above, 
				   size_t AUXSIM, // number of auxiliary datapoints between observations (capital M in the paper)
				   size_t DATASIZE, // number of actual datapoints (K+1 in the paper)
	   			   size_t CROSSSEC, // number of simulated paths (capital S in the paper)
				   double& a, // coefficient a as described in SDE above, result generated by EML will be written into a
				   double& b) // coefficient b as described SDE above, result generated by EML will be written into b
{
	size_t const INTERV(AUXSIM + 1); // number of intervals generated by auxiliary points between observations
	size_t const BIGDATASIZE(((DATASIZE - 1) * INTERV + 1) * CROSSSEC); // number of datapoints including auxiliaries
	double const DT(delta / static_cast<double>(INTERV)); // time interval between auxiliary datapoints
	double const ONEOVERDT(1.0 / DT);
	double const SQRTDT(sqrt(DT));
	double const ONEOVERSIGMA(1.0 / sigma);

	double* driftmod = new double[AUXSIM]; // will contain drift modification for modified brownian bridge from Durham and Gallant (2002)
	double* diffmod = new double[AUXSIM]; // will contain diffusion modification for modified brownian bridge from Durham and Gallant (2002)
	double* diffvar = new double[AUXSIM]; // will contain diffusion modification for modified brownian bridge from Durham and Gallant (2002)
	double* bridge = new double[BIGDATASIZE]; // will contain a cross section of the bridge process

	// fills the time-dependent coefficient of the modified Brownian bridge difference equation (eq 2.8 in the paper)
	for (size_t i = 0; i < AUXSIM; ++i){
			driftmod[i] = 1.0 / static_cast<double>(AUXSIM - i + 1);
			diffmod[i] = sqrt(static_cast<double>(AUXSIM - i) / static_cast<double>(AUXSIM - i + 1)) * SQRTDT;
			diffvar[i] =  1.0 / (diffmod[i] * diffmod[i]);
	}

	

	/************************
	*    GENERATE BRIDGE	*
	************************/

	// below we generate a cross section of the transformed bridge process
	// first of all we need to transform the data so that the transformed diffusion
	// has a unit diffusion function. The function that does the trick for the
	// square-root process is f:x->2 sqrt(x) / sigma
	// the transformed diffusion has dynamics
	// df = (c / f + e f) dt + dW, where c and e are related to a and b via
	// c = (2 a / sigma - 1/2), e = b / 2.
	// we will estimate c and e via EML and then transform them back into a and b
	// below we simulate the modified Brownian bridge from Durham and Gallant (2002)
	// note that the distribution of the discretized modified Bridge variates corresponds to the
	// exact distribution of a common diffusion bridge (Stramer and Yan (2007))
	
	// what might not be clear form the code below is that we generate a crosssection
	// of paths. All paths are augmented with auxiliary state variables from a bridge
	// Every path goes through the observed data points

	double bri0; // will contain backward variable for bridge
	double bri1; // forward variable
	double britr0; // transformed backward variable
	double britr1; // transformed forward variable
	size_t indBackward; // index helper variable
	size_t indForward; // index helper variable, unfortunately indexing is a real nightmare since we have so many dimensions (CROSSSED, AUXSIM, etc)
	size_t smallIndBackward; 
	size_t smallIndForward; 

	for (size_t i = 1; i < DATASIZE; ++i) {
		bri1 = data[i];
		bri0 = data[i - 1];
		britr1 = 2 * sqrt(bri1) * ONEOVERSIGMA;
		britr0 = 2 * sqrt(bri0) * ONEOVERSIGMA;
		indBackward = (i-1) * INTERV * CROSSSEC;
		indForward 	= i * INTERV * CROSSSEC;

		// when using Intel compiler the below loop can be vectorized
#pragma ivdep
		for (size_t k = 0; k < CROSSSEC; ++k){
			bridge[indBackward + k] = britr0;
		}
		for (size_t j = 1; j <= AUXSIM; ++j){
			smallIndBackward = indBackward + (j-1) * CROSSSEC;
			smallIndForward = indBackward + j * CROSSSEC;
#pragma ivdep
			for (size_t k = 0; k < CROSSSEC; ++k){
				bridge[smallIndForward + k] = bridge[smallIndBackward + k] + (britr1 - bridge[smallIndBackward + k]) * driftmod[j-1]
												+ diffmod[j-1] * randomNumbers[smallIndForward + k];
			}
		}
    }

	// very last value
	bri1 = data[DATASIZE - 1];
	britr1 = 2 * sqrt(bri1) * ONEOVERSIGMA;
	#pragma ivdep
	for (size_t k = 0; k < CROSSSEC; ++k){
		bridge[(DATASIZE-1) * INTERV * CROSSSEC + k] = britr1;
	}

	/***********
	*    EML   *
	***********/

	// This is the actual EML part. Please consult Mijatovic and Schneider (2007) for more information
	// The below loop Monte-Carlo integrates the conditional
	// expectations to obtain matrix A and vector b from the
	// paper (Mijatovic and Schneider (2007))
	// matrix A will be 2 x 2 symmetric positive semi-definite where the elements are computed over
	//
	//         1/f_t^2			1
	//
	//		   1				f_t^2

	// b will be of the form
	// 			(f_{t+1} - f_t) / f_t
	//			(f_{t+1} - f_t) * f_t

	// since A is symmetric semi-positive definite it will suffice to work with 3 elements
	// A_11, A_21, and A_22

	double A_11(0);
	double A_21(0);
	double A_22(0);

	double A_11c(0);
	double A_21c(0);
	double A_22c(0);

	double b_1(0);
	double b_2(0);

	double b_1c(0);
	double b_2c(0);

	const size_t BIGGUY((DATASIZE - 1) * INTERV + 1);
	double const SCALER(1.0 / static_cast<double>(CROSSSEC));

	for (size_t i = 1; i < BIGGUY; ++i){
		indBackward = (i-1) * CROSSSEC;
		indForward = i * CROSSSEC;
		A_11c = 0;
		A_21c = 0;
		A_22c = 0;
		b_1c = 0;
		b_2c = 0;
#pragma ivdep
		for (size_t k = 0; k < CROSSSEC; ++k){
			A_11c += 1.0 / (bridge[indBackward + k] * bridge[indBackward + k]);
			A_21c += 1.0 ;
			A_22c += (bridge[indBackward + k] * bridge[indBackward + k]);
			b_1c += (bridge[indForward + k] - bridge[indBackward + k]) / bridge[indBackward + k];
			b_2c += (bridge[indForward + k] - bridge[indBackward + k]) * bridge[indBackward + k];
		}
		A_11 += A_11c * SCALER * DT;
		A_21 += A_21c * SCALER * DT;
		A_22 += A_22c * SCALER * DT;
		b_1 += b_1c * SCALER;
		b_2 += b_2c * SCALER;
	}

	double A[2][2];
	A[0][0] = A_11;
	A[0][1] = A_21;
	A[1][0] = A_21;
	A[1][1] = A_22;

	double Ainv[2][2];
	EML::invertSymmPosMat<2>(Ainv, A);

	// do matrix multiplication, we get estimates for coefficients c and e as described above
	double c(Ainv[0][0] * b_1 + Ainv[0][1] * b_2 );
	double e(Ainv[1][0] * b_1 + Ainv[1][1] * b_2 );


	
	// now we can transform back into the coefficients we are interested in ( a and b)
	a = (c + 0.5) * 0.5 * sigma * sigma;
	b = e * 2;

	/************************
	*   Simulated Density	*
	************************/

	// here we compute an approximation to the log likelihood using simulated density
	// using Durham and Gallant (2002) importance sampler. Note that we can
	// *reuse* the cross section of bridges 

	double smalldens;
	double logdens;
	double* collect_q = new double[CROSSSEC];
	double* collect_p = new double[CROSSSEC];

	double x_mu_p;
	double x_mu_q;
	

    for (size_t i = 1; i < DATASIZE; ++i) {

		bri1 = data[i];
		bri0 = data[i - 1];
		britr1 = 2 * sqrt(bri1) * ONEOVERSIGMA;
		britr0 = 2 * sqrt(bri0) * ONEOVERSIGMA;
		indBackward = (i-1) * INTERV * CROSSSEC;
		indForward 	= i * INTERV * CROSSSEC;

		smalldens = 0;
		indBackward = (i-1) * INTERV * CROSSSEC;
		indForward 	= i * INTERV * CROSSSEC; 
#pragma ivdep
		for (size_t k = 0; k < CROSSSEC; ++k){
			collect_q[k] = 0;
			collect_p[k] = 0;
		}
		for (size_t j = 1; j <= AUXSIM; ++j){
			smallIndBackward = indBackward + (j-1) * CROSSSEC;
			smallIndForward = indBackward + j * CROSSSEC;
#pragma ivdep
			for (size_t k = 0; k < CROSSSEC; ++k){
				// here we compute f - mu(f) for the transformed square-root diffusion
				x_mu_p = bridge[smallIndForward + k] - bridge[smallIndBackward + k]
						- (c / bridge[smallIndBackward + k]  + e * bridge[smallIndBackward + k] ) * DT;

				// here we compute f-mu_{bridge}(f) for the modified bridge
				x_mu_q = bridge[smallIndForward + k] - bridge[smallIndBackward + k]
						- (britr1 - bridge[smallIndBackward + k]) * driftmod[j-1];
				
				
				// this is the kernel of a normal density, we take care of 1/2 and the minus sign below
				collect_p[k] += x_mu_p * x_mu_p * ONEOVERDT;
				collect_q[k] += x_mu_q * x_mu_q * diffvar[j-1];
			}
		}
#pragma ivdep
		for (size_t k = 0; k < CROSSSEC; ++k){
			// this is not a mistake. this is the transition density from the last auxiliary to the next observed point
			// there is no bridge density here purposely
			x_mu_p = (britr1 - bridge[indForward - CROSSSEC + k])
						- (c / bridge[indForward - CROSSSEC + k]  + e * bridge[indForward - CROSSSEC + k]) * DT;

			
			
			collect_p[k] += x_mu_p * x_mu_p * ONEOVERDT;
			
			smalldens += exp(-0.5 * (collect_p[k] - collect_q[k]));
		}
		// monte-carlo estimate of the transition density
		// plus change of variable accounting for the fact that we are looking at f=2 sqrt(x)/sigma
		// and not x directly ( remember change of variables formula p_x(x_i | x_{i-1}) = p_y(y_i | y_{i-1}) dy/dx 
		logdens += log(std::max(SCALER * smalldens, 1e-8))  - log(sigma * sqrt(bri1));
    }

	delete[] driftmod; 
	delete[] diffmod;
	delete[] diffvar;
	delete[] bridge;
	delete[] collect_p;
	delete[] collect_q;
	
	return logdens;
}

void emlTryOut(){
	
	const size_t DATASIZE(5041); // number of datapoints in the sample
	const size_t AUXSIM(23); // number of auxiliaries simulated between each datapoint
	const size_t CROSSSEC(AUXSIM * AUXSIM); // here we follow Stramer and Yan (2007) and set the number of paths in relation to AUXSIM
	const size_t BIGDATASIZE(((DATASIZE-1) * (AUXSIM + 1) + 1) * CROSSSEC); // I spent quite some time to figure out how large this array has to be
	const double DELTA(1.0 / 252.0);
	// this is VXO variance for 2525 days. VXO implied volatility is divided by 100 and squared
	std::ifstream inFile("varfile.txt");
	double* data = new double[DATASIZE];
	double* randomNumbers = new double[BIGDATASIZE];
	EML::fillWithNormalVariates(randomNumbers, BIGDATASIZE);

	// read in file
	for (size_t i = 0; i < DATASIZE; ++i){
		inFile >> data[i];
	}
	std::cout << "This function computes drift coefficients of a square-root diffusion:" << std::endl;
	std::cout << '\t' << "dx=(a+b x_t)dt + sigma sqrt(x_t)dW_t" << std::endl;
	std::cout << " input something reasonable for the diffusion coefficient (something in the range (0.001 - 1):" << std::endl;
	double mySigma;
	std::cin >> mySigma;
	if (mySigma < 0.00001 || mySigma > 1.0 ){
		std::cout << " Come on, take something more reasonable" << std::endl;
		exit(1);
	}

	// call eml function
	double a;
	double b;
	double propLikelihood = CIRestimate(randomNumbers, // vector (of size ((DATASIZE-1) * (AUXSIM + 1) + 1) * CROSSSEC) (containing N(0,1) random numbers
				  data, // vector (of size DATASIZE) containing the data
				  DELTA, // time between observations in vector data (assumed constant for simpl.)
				  mySigma, // value for diffusion coefficient in the SDE above
				  AUXSIM, // number of auxiliary datapoints between observations
				  DATASIZE, // number of actual datapoints
	   			  CROSSSEC, // how many paths do we simulate
				  a, // coefficient a as described in SDE above, result generated by EML will be written into a
				  b);

	std::cout << "parameter a: " << a << " parameter b: " << b << " proportional to likelihood: " << propLikelihood << std::endl;

	delete[] data;
	delete[] randomNumbers;
}


int main(int argc, char *argv[])
{
	// initialize C random generator, use fixed seed instead of time(0)
	// to reproduce results
	srand(time(0));

	
  	emlTryOut();

  return EXIT_SUCCESS;
}
