/***************************************************************************
 *   Copyright (C) 2008 by Paul Schneider   *
 *   paul.schneider@wu-wien.ac.at   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <cmath>
#include <cstdlib>

#ifndef M_PI
#define M_PI (3.14159265358979323846)
#endif

namespace EML {



// cholesky factorization A A' = M
// M symmetric positive definite
// result matrix is lower triangular matrix A


template<size_t dim>
inline		int choleskyFacSmall(double A[dim][dim],
                             	   const double M[dim][dim] ) { 
    double sum;
    for (size_t i = 0; i < dim; i++) {
        for (size_t j = i; j < dim; j++) {
            A[i][j] = M[i][j];
            sum = A[i][j];
            for (int k=i-1; k>= 0; k--) {
                // 				std::cout << k << std::endl;
                sum -= A[i][k] * A[j][k];
            }
            if(i == j) {
                if (sum <= 0) {
                    return EXIT_FAILURE;
                }
                A[i][i] = sqrt(sum);
            } else {
                A[j][i] = sum / A[i][i];
                A[i][j] = 0; 
            }
        }
    }
    return EXIT_SUCCESS;
}

// inverts lower triangular matrix M
// result is written into A

template<size_t dim>
inline		int invertLowerTriangular(double A[dim][dim],
                             	   const double M[dim][dim] ) {
    double sum(0);
    for (size_t i = 0; i < dim; i++) {
        A[i][i] = 1.0 / M[i][i];
        for (size_t j = i+1; j < dim; j++) {
            sum = 0;
            for (size_t k = i; k < j; ++k) {
                sum -= M[j][k]*A[k][i];
                A[j][i] = sum / M[j][j];
            }
        }
    }
    return EXIT_SUCCESS;
}


// computes the matrix product M' M
// result is written into A

template<size_t dim>
inline		int MTMProd(double A[dim][dim],
                        const double M[dim][dim] ) {
							
	double sum;
	for (size_t i = 0; i < dim; i++) {
        for (size_t j = i; j < dim; j++) {
            sum = 0;
            for (size_t k = j; k < dim; ++k) {
				sum += M[k][i]*M[k][j];
        	}
			A[i][j] = sum;
			A[j][i] = sum;  
    	}
	}
	return EXIT_SUCCESS;
}


// inverts symmetric positive semi-definite matrix M
// result is written into matrix A
template<size_t dim>
inline		int invertSymmPosMat(double A[dim][dim],
                             	   const double M[dim][dim] ) {
    double interMat[dim][dim];
    choleskyFacSmall<dim>(A, M);
    invertLowerTriangular<dim>(interMat, A);
    MTMProd<dim>(A, interMat);
    return EXIT_SUCCESS;
}


// fills a vector of double with size number of elements
// with N(0,1) variates. Use something more efficient than
// below in a real life app. Candidates are Gnu GSL or the
// vector random generators from the Intel Math Kernel Library

inline    void fillWithNormalVariates(double* vec, size_t size){
	
	static double const FACTOR   = 1.0 / (RAND_MAX);

	double u1;
	double u2;

	double x1;
	double x2;

	double inter;

	size_t counter = 0;

	while (counter < size-1){
		u1 = static_cast<double>(std::max(rand(), 1)) * FACTOR;
		u2 = static_cast<double>(std::max(rand(), 1)) * FACTOR;

		inter = sqrt(-2.0 * log(u1));

		x1 = inter * cos(2 * M_PI * u2);
		x2 = inter * sin(2 * M_PI * u2);

		vec[counter] = x1;
		vec[counter + 1] = x2;
		counter += 2;
	}
	if (size % 2){ // size uneven, we fill the last number
		u1 = static_cast<double>(std::max(rand(), 1)) * FACTOR;
		u2 = static_cast<double>(std::max(rand(), 1)) * FACTOR;

		inter = sqrt(-2.0 * log(u1));

		x1 = inter * cos(2 * M_PI * u2);
		vec[size-1] = x1;
	}

	
	
}

}

